use enum_map::{Enum, EnumMap};
use rand::{
    Rng,
    seq::SliceRandom,
    distributions::{Distribution, Uniform},
};

use crate::stack::Stack;


#[derive(Debug, Copy, Clone, Eq, PartialEq, Default, Enum)]
pub enum Color {
    #[default] Red, Green, Blue, Yellow, Purple,
}

// const COLORS: Stack<Color, 5> = Stack::from_array([
//     Color::Red,
//     Color::Green,
//     Color::Blue,
//     Color::Yellow,
//     Color::Purple,
// ]);

const COLORS: [Color; 5] = [
    Color::Red,
    Color::Green,
    Color::Blue,
    Color::Yellow,
    Color::Purple,
];


type ColorStack = Stack<Color, 5>;


#[derive(Debug, Copy, Clone)]
pub enum Tile {
    Forward,
    Backward,
}


#[derive(Debug, Copy, Clone)]
pub enum Square {
    Camels(ColorStack),
    Tile(Tile),
}

impl Square {
    fn assume_stack(&self) -> &ColorStack {
        match self {
            Square::Camels(stack) => stack,
            _ => panic!("Attempted to use the stack from a non-stack square"),
        }
    }

    fn assume_stack_mut(&mut self) -> &mut ColorStack {
        match self {
            Square::Camels(stack) => stack,
            _ => panic!("Attempted to use the stack from a non-stack square"),
        }
    }
}

impl Default for Square {
    fn default() -> Self {
        Square::Camels(ColorStack::new())
    }
}


#[derive(Debug, Default)]
pub struct Game {
    squares: [Square; 16],
    dice: EnumMap<Color, bool>,
    camels: EnumMap<Color, usize>,
}

impl Game {
    fn new() -> Self {
        Self::default()
    }

    // new game with random starting positions
    fn new_random() -> Self {
        let mut game = Self::default();
        let mut rng = rand::thread_rng();
        let mut dice = *&COLORS;
        dice.shuffle(&mut rng);
        for color in dice {
            let roll = rng.gen_range(1..=3);
            game.squares[roll - 1].assume_stack_mut().push(color);
            game.camels[color] = roll - 1;
        }
        game
    }

    fn set_state(&mut self, squares: [Square; 16], dice: EnumMap<Color, bool>) {
        self.squares = squares;
        self.dice = dice;
        for (i, square) in self.squares.iter().enumerate() {
            if let Square::Camels(stack) = square {
                for camel in stack.iter() {
                    self.camels[*camel] = i
                }
            }
        }
    }

    // returns winner if there is one
    fn advance(&mut self, die: Color, roll: usize) -> Option<Color> {
        let src_sq = self.camels[die];
        let dst_sq = src_sq + roll;
        if dst_sq >= 16 {
            self.dice[die] = true;
            return self.squares[src_sq].assume_stack().last().copied();
        }

        // special case when the destination square is the same as the source square
        if let Square::Tile(Tile::Backward) = self.squares[dst_sq] {
            if roll == 1 {
                let src_stack = self.squares[src_sq].assume_stack_mut();
                let slice_start = src_stack.iter().position(|&c| c == die).unwrap();
                src_stack.shift_slice_under(slice_start);
            }
        }
        else {
            // we have to split self.squares into two slices using split_at_mut, otherwise
            // rustc complains that we're trying to use two mutable references to the same value
            let (left, right) = self.squares.split_at_mut(src_sq + 1);
            let src_stack = left[src_sq].assume_stack_mut();
            let slice_start = src_stack.iter().position(|&c| c == die).unwrap();

            // since `right` starts immediately after the source square, the index of the 
            // destination square will be roll - 1 (e.g. if roll is 1, dst will be right[0])
            let (dst_rel_idx, prepend) = match right[roll - 1] {
                Square::Tile(Tile::Forward) => (roll, false), // roll - 1 + 1
                Square::Tile(Tile::Backward) => (roll - 2, true), // roll is guaranteed to be >= 2 since we already handled roll == 1
                _ => (roll - 1, false),
            };
            let dst_stack = right[dst_rel_idx].assume_stack_mut();
            let dst_true_idx = src_sq + 1 + dst_rel_idx; // src_sq + 1 was the original split boundary, so add the relative index to that to get the true index

            if prepend {
                let slice_len = src_stack.len() - slice_start;
                src_stack.move_slice_under(dst_stack, slice_start);
                for i in 0..slice_len {
                    self.camels[dst_stack[i]] = dst_true_idx;
                }
            }
            else {
                let dst_prev_len = dst_stack.len();
                src_stack.move_slice(dst_stack, slice_start);
                for i in dst_prev_len..dst_stack.len() {
                    self.camels[dst_stack[i]] = dst_true_idx;
                }
            }
        }

        self.dice[die] = true;
        None
    }

    fn finish_leg_random(&mut self) -> Option<Color> {
        let mut rng = rand::thread_rng();
        let mut leg_dice: Stack<Color, 5> = Stack::new();
        for (color, rolled) in self.dice {
            if !rolled {
                leg_dice.push(color);
            }
        }

        (&mut leg_dice[..]).shuffle(&mut rng);

        for color in leg_dice.iter() {
            let roll = rng.gen_range(1..=3);
            if let Some(winner) = self.advance(*color, roll) {
                return Some(winner);
            }
        }
        None
    }

    fn finish_game_random(&mut self) -> Color {
        if let Some(winner) = self.finish_leg_random() {
            return winner;
        }

        // we are now guaranteed to be at the start of a new leg,
        // so we don't need to check the dice state
        let mut rng = rand::thread_rng();
        let roll_dist = Uniform::from(0..=3);
        let mut dice = *&COLORS; // makes a copy of the constant

        loop {
            // easiest if we shuffle at the start of the leg
            dice.shuffle(&mut rng);
            for i in 0..5 {
                let roll = roll_dist.sample(&mut rng);
                if let Some(winner) = self.advance(dice[i], roll) {
                    return winner;
                }
            }
        }
    }
}


#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn test_advance() {
        use Color::*;
        let mut game = Game::new();
        // all dice are false (not rolled) to start with
        assert_eq!(game.dice.values().any(|&v| v), false);
        
        let mut squares = [Square::Camels(Default::default()); 16];
        let one = squares[0].assume_stack_mut();
        one.push(Blue);
        one.push(Yellow);
        squares[1].assume_stack_mut().push(Red);
        let three = squares[2].assume_stack_mut();
        three.push(Green);
        three.push(Purple);
        //  BY, R, GP

        game.set_state(squares, Default::default());

        game.advance(Yellow, 2);
        println!("{:?}", game.camels);
        assert_eq!(game.dice[Yellow], true);
        assert_eq!(game.camels[Yellow], 2);
        assert_eq!(game.squares[2].assume_stack(), &Stack::from([Green, Purple, Yellow]));
        // B, R, GPY

        game.advance(Red, 2);
        assert_eq!(game.dice[Red], true);
        assert_eq!(game.camels[Red], 3);
        // B, _, GPY, R
        
        game.advance(Purple, 1);
        assert_eq!(game.dice[Purple], true);
        assert_eq!(game.squares[3].assume_stack(), &Stack::from([Red, Purple, Yellow]));
        // B, _, G, RPY
    }


}
